<?php

namespace Drupal\monarch_media_bulk_upload\Form;

use Drupal\Core\Entity\Display\EntityFormDisplayInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Render\Element;
use Drupal\filter\Render\FilteredMarkup;
use Drupal\media\MediaInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a Monarch Media Bulk Upload form.
 */
class BulkUpdateForm extends FormBase {

  use MessengerTrait;

  /**
   * The entity_type.manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    $instance = parent::create($container);

    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->currentUser = $container->get('current_user');
    $instance->formBuilder = $container->get('form_builder');
    $instance->configFactory = $container->get('config.factory');

    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'monarch_media_bulk_upload_bulk_update';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $idlist = NULL) {
    unset($form['#after_build']);

    $form['#tree'] = TRUE;
    $form['#parents'] = $form['#parents'] ?? [];

    $config = $this->configFactory->get('monarch_media_bulk_upload.settings');
    if ($page_class = $config->get('page_class')) {
      $form['#attributes']['class'][] = $page_class;
    }

    if (is_string($idlist)) {
      $idlist = array_filter(preg_split("/[^\d]+/", $idlist));

      if (!empty($idlist)) {
        $ms = $this->entityTypeManager->getStorage('media');
        $efd = $this->entityTypeManager->getStorage('entity_form_display');

        /** @var \Drupal\media\MediaInterface[] $medias */
        $medias = array_filter($ms->loadMultiple($idlist), function (MediaInterface $media) use ($efd) {
          if (!$media->access('update', $this->currentUser())) {
            return FALSE;
          }

          $display = $efd->load('media.' . $media->bundle() . '.bulk_upload');
          return $display instanceof EntityFormDisplayInterface && !empty($display) && $display->get('status');
        });

        if (!empty($medias)) {
          $form['#parents'] = $form['#parents'] ?? [];
          $form['media'] = [
            '#parents' => [...$form['#parents'], 'media'],
            '#type' => 'table_container',
            '#header' => [],
          ];

          $field_list = [];

          foreach ($medias as $media) {
            foreach ($this->getWidgets($media) as $name => $widget) {
              $field_list[$name] = $name;

              if ($media->hasField($name)) {
                $form['media']['#header'][$name] = $media->get($name)->getFieldDefinition()->getLabel();
              }
              else {
                $form['media']['#header'][$name] = $form['media']['#header'][$name] ?? $name;
              }
            }
          }

          foreach ($medias as $media) {
            $form['media'][$media->id()]['#parents'] = [
              ...$form['media']['#parents'],
              $media->id(),
            ];

            foreach ($field_list as $name) {
              $form['media'][$media->id()][$name] = [
                '#markup' => '&nbsp;',
              ];
            }

            $this->buildSubform($media, $form['media'][$media->id()], $form_state);
          }
        }
      }
    }

    $form['actions'] = [
      '#type' => 'actions',
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Submit'),
        '#submit' => ['::submitForm', '::save'],
      ],
    ];

    return $form;
  }

  /**
   * Hides title for the element.
   */
  protected function hideTitle(array &$element) {
    if (!is_null($element)) {
      if ($element['#title_display'] ?? NULL) {
        $element['#title_display'] = 'hidden';
      }
      elseif ($element['#title'] ?? NULL) {
        unset($element['#title']);
      }
    }
  }

  /**
   * Gets the subform.
   */
  protected function buildSubform(FieldableEntityInterface $entity, array &$form, FormState $form_state) {
    $form['#parents'] = $form['#parents'] ?? [];
    $form['#entity'] = $entity;

    foreach ($this->getWidgets($entity) as $name => $widget) {
      if ($this->getComponent($entity, $name)) {
        $items = $entity->get($name);
        $items->filterEmptyItems();
        $form[$name] = $widget->form($items, $form, $form_state);
        $form[$name]['#access'] = $items->access('edit');

        if ($form[$name] ?? NULL) {
          $this->hideTitle($form[$name]);
        }

        if ($form[$name]['widget'] ?? NULL) {
          $this->hideTitle($form[$name]['widget']);
        }

        for ($i = 0; $form[$name][$i] ?? NULL; $i++) {
          $this->hideTitle($form[$name][$i]);

          if ($form[$name][$i]['value'] ?? NULL) {
            $this->hideTitle($form[$name][$i]['value']);
          }
        }

        for ($i = 0; $form[$name]['widget'][$i] ?? NULL; $i++) {
          $this->hideTitle($form[$name]['widget'][$i]);

          if ($form[$name]['widget'][$i]['value'] ?? NULL) {
            $this->hideTitle($form[$name]['widget'][$i]['value']);
          }
        }
      }
    }
  }

  /**
   * Get entity widget.
   *
   * @return \Drupal\Core\Field\WidgetInterface
   *   The renderer.
   */
  protected function getRenderer(FieldableEntityInterface $entity, string $name) {
    $efd = $this->entityTypeManager->getStorage('entity_form_display');
    /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display */
    $display = $efd->load('media.' . $entity->bundle() . '.bulk_upload');
    return $display->getRenderer($name);
  }

  /**
   * Get entity widget.
   *
   * @return array|null
   *   The renderer.
   */
  protected function getComponent(FieldableEntityInterface $entity, string $name) : ?array {
    $efd = $this->entityTypeManager->getStorage('entity_form_display');
    /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display */
    $display = $efd->load('media.' . $entity->bundle() . '.bulk_upload');
    return $display->getComponent($name);
  }

  /**
   * Get entity widgets.
   *
   * @return \Drupal\Core\Field\WidgetInterface[]
   *   The array of widgets.
   */
  protected function getWidgets(FieldableEntityInterface $entity) : array {
    $efd = $this->entityTypeManager->getStorage('entity_form_display');
    /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display */
    $display = $efd->load('media.' . $entity->bundle() . '.bulk_upload');

    $widgets = [];

    if ($display) {
      foreach ($display->getComponents() as $name => $options) {
        if ($name !== 'revision_log_message') {
          if ($widget = $display->getRenderer($name)) {
            $widgets[$name] = [$widget, $options['weight'] ?? NULL ?: 0];
          }
        }
      }
    }

    uasort($widgets, function ($a, $b) {
      return $a[1] - $b[1];
    });

    return array_map(function ($entry) {
      return $entry[0];
    }, $widgets);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    foreach (Element::children($form['media']) as $entity_id) {
      if ($form['media'][$entity_id]['#entity'] ?? NULL) {
        $this->extractFormValues($form['media'][$entity_id]['#entity'], $form['media'][$entity_id], $form_state);
        $this->validateFormValues($form['media'][$entity_id]['#entity'], $form['media'][$entity_id], $form_state);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function extractFormValues(FieldableEntityInterface $entity, array &$form, FormStateInterface $form_state) {
    $efd = $this->entityTypeManager->getStorage('entity_form_display');
    /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display */
    $display = $efd->load('media.' . $entity->bundle() . '.bulk_upload');
    return $display->extractFormValues($entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateFormValues(FieldableEntityInterface $entity, array &$form, FormStateInterface $form_state) {
    $efd = $this->entityTypeManager->getStorage('entity_form_display');
    /** @var \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display */
    $display = $efd->load('media.' . $entity->bundle() . '.bulk_upload');
    return $display->validateFormValues($entity, $form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $form_state->setRedirect('media_bulk_upload.list');
    $form_state->cleanValues();

    $media_links = [];

    foreach (Element::children($form['media']) as $entity_id) {
      if ($form['media'][$entity_id]['#entity'] ?? NULL) {
        $entity = $form['media'][$entity_id]['#entity'] = $this->buildEntity($form['media'][$entity_id]['#entity'], $form['media'][$entity_id], $form_state);
        $media_links[] = $entity->toLink()->toString();
      }
    }

    if (!empty($media_links)) {
      $this->messenger()->addMessage(FilteredMarkup::create('Updated media: ' . implode(', ', $media_links)));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function save(array &$form, FormStateInterface $form_state) {
    foreach (Element::children($form['media']) as $entity_id) {
      if ($form['media'][$entity_id]['#entity'] ?? NULL) {
        $form['media'][$entity_id]['#entity']->save();
      }
    }
  }

  /**
   * Builds an updated entity object based upon the submitted form values.
   *
   * For building the updated entity object the form's entity is cloned and
   * the submitted form values are copied to entity properties. The form's
   * entity remains unchanged.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $source_entity
   *   The entity to rebuild using form values.
   * @param array $form
   *   A nested array form elements comprising the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Entity\EntityInterface
   *   An updated copy of the form's entity object.
   *
   * @see \Drupal\Core\Entity\EntityFormInterface::getEntity()
   */
  public function buildEntity(FieldableEntityInterface $source_entity, array $form, FormStateInterface $form_state) {
    $entity = clone $source_entity;
    $this->copyFormValuesToEntity($entity, $form, $form_state);

    // Invoke all specified builders for copying form values to entity
    // properties.
    if (isset($form['#entity_builders'])) {
      foreach ($form['#entity_builders'] as $function) {
        call_user_func_array($form_state->prepareCallback($function), [
          $entity->getEntityTypeId(),
          $entity,
          &$form,
          &$form_state,
        ]);
      }
    }

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(FieldableEntityInterface $entity, array $form, FormStateInterface $form_state) {
    // First, extract values from widgets.
    $extracted = $this->extractFormValues($entity, $form, $form_state);

    // Then extract the values of fields that are not rendered through widgets,
    // by simply copying from top-level form values. This leaves the fields
    // that are not being edited within this form untouched.
    foreach ($form_state->getValues() as $name => $values) {
      if ($entity->hasField($name) && !isset($extracted[$name])) {
        $entity->set($name, $values);
      }
    }
  }

}
